from typing import Any

from confme.parsing.parser_secret import ParserSecret
from parsing.parser_primitive import ParserPrimitive


class UnknownTypeException(Exception):
    pass


def parse(attr_name: str, attr_type: Any, value: Any):
    parser = [p for p in PARSER if p.is_applicable(attr_name, attr_type)]
    if len(parser) > 1:
        raise Exception('More than one Parser present to parse the same attribute type')
    if len(parser) < 1:
        raise UnknownTypeException()

    return parser[0].parse(attr_name, attr_type, value)


PARSER = [
    ParserPrimitive(parse),
    ParserSecret(parse)
]



import os
from typing import Any

from confme.parsing.parser_base import Parser, ParseError


class Secret:
    __slots__ = ()

    def __new__(cls, *args, **kwds):
        raise SyntaxError(f'{cls.__name__} is only a marker class and can not be instantiated')

    def __class_getitem__(cls, params):
        assert len(params) == 2, 'Params must be of length two'

        cls.__key__ = params[0]
        cls.__value_type__ = params[1]

        return cls

    def __init_subclass__(cls, *args, **kwargs):
        pass


class ParserSecret(Parser):

    def __init__(self, propagate_parse_cb):
        super(ParserSecret, self).__init__(propagate_parse_cb)

    def is_applicable(self, attr_name: str, attr_type: Any):
        return attr_type == Secret

    def parse(self, attr_name: str, attr_type: Secret, value: Any):
        if attr_type.__key__ not in os.environ:
            raise ParseError(f'Secret with name {attr_type.__key__} not found')

        return self.propagate_parse_cb(attr_name,
                                       attr_type.__value_type__,
                                       os.environ[attr_type.__key__])




import logging
from abc import abstractmethod
from os import path
from typing import Dict

import yaml
from yaml.parser import ParserError


class BaseFileParser:
    @abstractmethod
    def get_endings(self):
        pass

    @abstractmethod
    def parse(self, file: str):
        pass


class YamlFileParser(BaseFileParser):

    def get_endings(self):
        return ['.yaml', '.yml']

    def parse(self, file):
        try:
            return yaml.safe_load(file)
        except ParserError:
            logging.exception('Not able to parse yaml file')
            raise ParserError


FILE_PARSER = [
    YamlFileParser()
]


def parse_file(file_path: str) -> Dict:
    ending = path.splitext(file_path)[-1]
    applicable_file_parsers = [p for p in FILE_PARSER if ending in p.get_endings()]
    if len(applicable_file_parsers) < 0:
        raise Exception(f'File Ending {ending} not known')
    if len(applicable_file_parsers) > 1:
        raise Exception(f'More than one parser registered for this file ending... 🧐')

    with open(file_path, 'r') as file:
        return applicable_file_parsers[0].parse(file)


from typing import Union

from confme.parsing.parser_base import Parser, ParseError


class ParserPrimitive(Parser):
    SCALAR_TYPES = Union[int, float, str]

    def __init__(self, propagate_parse_cb):
        super(ParserPrimitive, self).__init__(propagate_parse_cb)

    def is_applicable(self, attr_name, attr_type):
        return self._check_union(attr_type, self.SCALAR_TYPES)

    def parse(self, attr_name, attr_type, value):
        try:
            return attr_type(value)
        except ValueError:
            raise ParseError()



from dataclasses import dataclass
from typing import get_type_hints, Dict

from file_backend import file_backend
from parsing import root_parser


def configclass(_cls):
    _cls.__config_class__ = True
    return dataclass(_cls)


def is_config_class(cls):
    if not getattr(cls, '__dict__', None):
        return False

    return '__config_class__' in cls.__dict__


def load_config(config_class, path: str):
    config_content = file_backend.parse_file(path)

    return fill_config_classes(config_class, config_content)


def fill_config_classes(config_type_head, yaml_head: Dict):
    params = {}
    for attr_name, attr_type in get_type_hints(config_type_head).items():
        if is_config_class(attr_type):
            params[attr_name] = fill_config_classes(attr_type, yaml_head[attr_name])
        else:
            params[attr_name] = root_parser.parse(attr_name, attr_type, yaml_head.get(attr_name, None))

    return config_type_head(**params)


from abc import abstractmethod, ABC


class ParseError(Exception):

    def __init__(self, msg: str=None):
        self.msg = msg


class Parser(ABC):

    def __init__(self, propagate_parse_cb):
        self.propagate_parse_cb = propagate_parse_cb

    @abstractmethod
    def is_applicable(self, attr_name, attr_type):
        pass

    def _check_union(self, type_to_check, union):
        return any([union_t == type_to_check for union_t in union.__args__])

    def parse(self, attr_name, attr_type, value):
        pass
